import gc
import weave
from hazm import word_tokenize as persian_word_tokenize,sent_tokenize as persian_sent_tokenize
from nltk.tokenize import word_tokenize as english_word_tokenize,sent_tokenize as english_sent_tokenize
import numpy as np
import re
import wandb
from collections import OrderedDict

has_logged_in_to_wandb = False




def init_wandb(project_name, run_name, config=None,group=None):
    if not Constants.use_wandb:
        return

    global has_logged_in_to_wandb
    if not has_logged_in_to_wandb:
        api_key = Constants.WandbToken
        wandb.login(key=api_key)
        has_logged_in_to_wandb = True

    wandb.init(project=project_name, name=run_name, config=config,group=group)


def init_hf():
    from huggingface_hub import login
    hf_token = Constants.HFToken
    login(token=hf_token, add_to_git_credential=False)


def listToString(listOfStrings, separator="\n"):
    output = ""
    for s in listOfStrings:
        output += str(s) + separator
    return output

def get_word_tokenizer():
    if Constants.lang=="multi":
        return english_word_tokenize
    elif Constants.lang=="fa":
        return persian_word_tokenize

    print("Error: lang not set to fa or multi")
    return None

def get_sent_tokenizer():
    if Constants.lang=="multi":
        return english_sent_tokenize
    elif Constants.lang=="fa":
        return persian_sent_tokenize

    print("Error: lang not set to fa or multi")
    return None


def capped_sent_tokenize(text, max_length=500):
    custom_word_tokenizer = get_word_tokenizer()
    sentence_tokenizer = get_sent_tokenizer()
    sentences = sentence_tokenizer(text)
    capped_sentences = []
    for sentence in sentences:
        while token_len(sentence) > max_length:
            tokens = custom_word_tokenizer(sentence)
            capped_sentences.append(" ".join(tokens[:max_length]))
            sentence = " ".join(tokens[max_length:])
        capped_sentences.append(sentence)
    return capped_sentences


def token_len(text):
    custom_word_tokenizer = get_word_tokenizer()
    return len(custom_word_tokenizer(text))




def reflect_vector(q, v):
    proj = np.dot(v, q) / np.dot(q, q) * q
    reflection = 2 * proj - v

    return reflection


def read_pdf(pdf_path, ignore_line_breaks=True):
    import PyPDF2

    text = ""
    with open(pdf_path, 'rb') as file:
        pdf_reader = PyPDF2.PdfReader(file)
        for page in pdf_reader.pages:
            if page.extract_text():
                page_text = page.extract_text()
                if ignore_line_breaks:
                    # Replace single line breaks with space
                    page_text = re.sub(r'(?<!\n)\n(?!\n)', ' ', page_text)
                    text += page_text + '\n'
                else:
                    text += page_text

    return text.strip()


def Init(encoder_name, top_k, sample_size, qa_sample_size,  multi_hop_hardness_factor=0,
         judged=False, useTool=False, llm=None,multi_hop=True,custom_data=None,custom_project_name=None,wandb_group=None):
    from MetaRagTool.Utils.MetaRagConfig import MetaRagConfig
    from MetaRagTool.Encoders.SentenceTransformerEncoder import SentenceTransformerEncoder
    from MetaRagTool.LLM.JudgeLLM import JudgeLLM
    from MetaRagTool.Utils import DataLoader
    from MetaRagTool.LLM.GoogleGemini import Gemini

    gc.collect()

    Constants.use_wandb = True

    if custom_data is not None:
        contexts, qas = custom_data
        multi_hop = False
    else:
        contexts, qas = DataLoader.loadWikiFaQa(sample_size=sample_size, qa_sample_size=qa_sample_size, multi_hop=multi_hop,
                                            multi_hop_hardness_factor=multi_hop_hardness_factor)


    encoder = SentenceTransformerEncoder(encoder_name)
    judge = None

    llm_name= None
    if judged:
        if llm is None:
            llm = Gemini(has_memory=False, RequestPerMinute_limit=15, model_name=Gemini.GEMINI_2_FLASH)
            # llm = OpenaiGpt(has_memory=False, RequestPerMinute_limit=15)

        judge = JudgeLLM(model_name=Gemini.GEMINI_2_FLASH)
        llm_name=llm.model_name
        project_name = "fullRag"
        weave.init(project_name)
    else:
        project_name = 'retrival'

    if custom_project_name is not None:
        project_name = custom_project_name

    ragConfig = MetaRagConfig(encoder_name=encoder_name, top_k=top_k, sample_size=sample_size,
                              qa_sample_size=qa_sample_size,
                              multi_hop_hardness_factor=multi_hop_hardness_factor, judged=judged, useTool=useTool,
                              llm_name=llm_name,
                              encoder=encoder, contexts=contexts, qas=qas, llm=llm, judge=judge,
                              project_name=project_name,multi_hop=multi_hop,wandb_group=wandb_group)

    return ragConfig


def remove_duplicates(l):
    return list(OrderedDict.fromkeys(l))



import MetaRagTool.Constants as Constants



import MetaRagTool.Utils.DataLoader
import MetaRagTool.Utils.MRUtils

from MetaRagTool.Utils.MRUtils import init_hf
from MetaRagTool.Utils.MetaRagConfig import MetaRagConfig,get_best_config,get_baseline_config


# print("Utils imported")



use_wandb = False
local_mode = False
trust_remote_code = True
lang = 'multi'
mute=False

WandbToken = None
HFToken = None
GitHubToken = None

API_KEY_GEMINI = None
API_KEY_OPENAI = None
API_KEY_OPENROUTER = None

import nltk

try:
    nltk.data.find('tokenizers/punkt')
except BaseException:
    try:
        nltk.download('punkt')

    except BaseException as e:
        print(f"Error downloading 'punkt' for nltk: {e}")

try:
    nltk.data.find('tokenizers/punkt_tab')
except BaseException:
    try:
        nltk.download('punkt_tab')
    except BaseException as e:
        print(f"Error downloading 'punkt_tab' for nltk: {e}")



from MetaRagTool.Encoders.SentenceTransformerEncoder import SentenceTransformerEncoder
from MetaRagTool.Encoders.OtherEncoders import BertEncoder,XLMRobertaEncoder,AutoModelEncoder
from MetaRagTool.Encoders.Encoder import Encoder,OneByOneEncoder
from MetaRagTool.Encoders.Reranker import Reranker,AutoModelForSequenceClassificationReranker,CrossEncoderReranker

# print("Encoders imported")




import sentence_transformers
from sentence_transformers import SentenceTransformer
from MetaRagTool.Encoders.Encoder import Encoder


class SentenceTransformerEncoder(Encoder):
    MODEL_NAME_LABSE = "sentence-transformers/LaBSE"
    MODEL_NAME_E5SMALL='intfloat/multilingual-e5-small'
    def __init__(self, model_name: str, verbose=False,query_prompt=None,doc_prompt=None):
        import MetaRagTool.Constants as Constants

        super().__init__(model_name, verbose)
        if Constants.trust_remote_code:
            self.model = SentenceTransformer(model_name,trust_remote_code=True)
        else:
            self.model = SentenceTransformer(model_name)

        self.query_prompt= query_prompt
        self.doc_prompt=doc_prompt
        print("Model loaded successfully")

    def encode(self, sentences, isQuery=True):
        if sentence_transformers.__version__ >= "5.0.0":
            if isQuery:
                embeddings = self.model.encode_query(sentences,
                                               # batch_size=256,
                                               show_progress_bar=not isQuery,
                                               convert_to_tensor=False, normalize_embeddings=True,prompt=self.query_prompt)

            else:
                embeddings = self.model.encode_document(sentences,
                                               # batch_size=256,
                                               show_progress_bar=not isQuery,
                                               convert_to_tensor=False, normalize_embeddings=True,prompt=self.doc_prompt)


        else:

            prompt=self.doc_prompt
            if isQuery:
                prompt = self.query_prompt

            embeddings = self.model.encode(sentences,
                                                    # batch_size=256,
                                                    show_progress_bar=not isQuery,
                                                    convert_to_tensor=False, normalize_embeddings=True,prompt=prompt)


        return embeddings





from setuptools import setup, find_packages
import re
import os

# Function to extract the version from __init__.py
def get_version():
    init_py_path = os.path.join(os.path.dirname(__file__), 'MetaRagTool', '__init__.py')
    with open(init_py_path, 'r') as f:
        version_file_content = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file_content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

version = get_version()

setup(
    name='MetaRagTool',
    packages=find_packages(),
    version=version,
    license='MIT',
    description='A comprehensive RAG (Retrieval-Augmented Generation) toolkit for AI applications',
    author='Ali Mobarekati',
    author_email='ali7919mb@yahoo.com',
    url='https://github.com/ali7919/MetaRAG',
    download_url=f'https://github.com/ali7919/MetaRAG/archive/refs/tags/v{version}.tar.gz', # Dynamically set
    keywords=['AI', 'RAG', 'LLM', 'NLP', 'Machine Learning'],
    install_requires=[
        'faiss-cpu',
        'hazm',
        'PyPDF2',
        'datasets>=3',
        'huggingface_hub',
        'gradio',
        'weave',

        # you can remove these and it will still work on google Colab
        'transformers',
        'sentence-transformers>=5',
        'langchain-text-splitters',
        'matplotlib',
        'wandb==0.19.2',
        'google-genai',
        'openai',
        'tqdm',
        'nltk',
        'pandas',
        # 'numpy'

    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)


import gc
import math
import MetaRagTool.Constants as Constants
# from tqdm.notebook import tqdm
from tqdm import tqdm

import wandb
import time as time


import MetaRagTool.Utils.MRUtils as MRUtils
from MetaRagTool.RAG.MetaRAG import MetaRAG
from MetaRagTool.Utils.MetaRagConfig import MetaRagConfig
from MetaRagTool.Utils.MRUtils import token_len

verbose_run = False


from abc import ABC, abstractmethod
class EvaluatableRAG(ABC):
    def __init__(self):
        self.chunk_size= -1
        self.ChunksList=[]
        self.time_to_encode_corpus=-1

    @abstractmethod
    def retrieve(self,query, top_k):
        pass

    @abstractmethod
    def add_corpus(self,contexts:list):
        pass

def check_exact_retrieval(rag, pair, verbose=False, top_k=5):
    query = pair[1]['sentence_A']
    answer = pair[1]['sentence_B']

    retrieved = rag.retrieve(query, top_k=top_k)

    if verbose:
        print0(f'query : {query}\n answer : {answer}\n retrieved : ')
        print0(MRUtils.listToString(retrieved))

    if answer in retrieved:
        if verbose: print0('\n Correct')
        return 1
    else:
        if verbose: print0('\n Incorrect')
        return 0


def full_test_exact_match(rag, testData, top_k=5):
    correct = 0
    counter = 0

    tqdm_loop = tqdm(testData.iterrows(), total=len(testData), desc="Evaluating")
    if Constants.mute:tqdm_loop.disable=True

    for row in tqdm_loop:
        counter += 1
        correct += check_exact_retrieval(rag=rag, pair=row, top_k=top_k)
        tqdm_loop.set_postfix({'Accuracy': correct / counter})

    print0(f'{correct} out of {counter} Accuracy = {correct / counter}')


def full_test_find_in(rag,ragConfig:MetaRagConfig, question_key='question', answer_key='answer',verbose=False):
    t1 = time.time()
    correct = 0
    counter = 0
    average_precision = 0

    all_retrieved_length = 0
    correct_retrieved_length = 0

    tqdm_loop = tqdm(ragConfig.qas.iterrows(), total=len(ragConfig.qas), desc="Evaluating",disable=not ragConfig.fine_grain_progressbar)
    if Constants.mute:tqdm_loop.disable=True

    for row in tqdm_loop:
        counter += 1
        query = row[1][question_key]
        answer = row[1][answer_key]

        retrieved = rag.retrieve(query, top_k=ragConfig.top_k)
        retrieved = MRUtils.listToString(retrieved)

        retrieved_len = token_len(retrieved)
        all_retrieved_length += retrieved_len

        if answer in retrieved:
            correct += 1
            correct_retrieved_length += retrieved_len
            average_precision += token_len(answer) / retrieved_len

        if verbose:
            print0(f'query : {query}\n answer : {answer}\n retrieved : ')
            print0(retrieved)

        acc = correct / counter
        avg_len = math.floor(all_retrieved_length / counter)
        efficiency = math.floor((acc * 100) ** 3 / avg_len)

        tqdm_loop.set_postfix({'Accuracy': acc, 'Avg retrieved length': str(avg_len), 'Efficiency': str(efficiency)})

    total_time = time.time() - t1
    average_precision /= counter
    print0(f'Correct : {correct} out of {counter} = {acc} ')
    print0(f'Average Precision : {average_precision}')
    print0(f'Length Accuracy Efficiency : {efficiency} ')
    print0(f'Average retrieved length : {avg_len}')
    print0(f'Total time : {total_time}')

    if Constants.use_wandb:
        wandb.log({
            'k': ragConfig.top_k,
            'chunk_size': rag.chunk_size,
            'Accuracy': acc,
            'Avg retrieved length': avg_len,
            'Efficiency': efficiency,
            'Average Precision': average_precision,
            'Total time': total_time,
            'corpus size': len(rag.ChunksList),
            'time to encode corpus': rag.time_to_encode_corpus

        })

    return {
        "avg_len": avg_len,
        "efficiency": efficiency,
        "accuracy": acc
    }

def full_test_find_in_multiHop(rag,ragConfig:MetaRagConfig,verbose =False):
    t1 = time.time()
    foundBoth = 0
    foundOne = 0
    counter = 0
    average_precision = 0

    all_retrieved_length = 0
    correct_retrieved_length = 0

    tqdm_loop = tqdm(ragConfig.qas.iterrows(), total=len(ragConfig.qas), desc="Evaluating",disable=not ragConfig.fine_grain_progressbar)
    if Constants.mute:tqdm_loop.disable=True
    for row in tqdm_loop:
        counter += 1
        query = row[1]['question']
        answer1 = row[1]['answer1']
        answer2 = row[1]['answer2']

        retrieved = rag.retrieve(query, top_k=ragConfig.top_k)
        retrieved = MRUtils.listToString(retrieved)

        retrieved_len = token_len(retrieved)

        all_retrieved_length += retrieved_len

        if answer1 in retrieved and answer2 in retrieved:
            foundBoth += 1
            foundOne += 1
            correct_retrieved_length += retrieved_len
        elif answer1 in retrieved or answer2 in retrieved:
            foundOne += 1

        if verbose:
            print0(f'query : {query}\n answer1 : {answer1}\n answer2 : {answer2}\n retrieved : ')
            print0(retrieved)

        full_acc = foundBoth / counter
        half_acc = foundOne / counter

        if answer1 in retrieved and answer2 in retrieved:
            average_precision += (token_len(answer1) + token_len(answer2)) / retrieved_len
        elif answer1 in retrieved or answer2 in retrieved:
            if answer1 in retrieved:
                average_precision += token_len(answer1) / retrieved_len
            if answer2 in retrieved:
                average_precision += token_len(answer2) / retrieved_len

        avg_len = math.floor(all_retrieved_length / counter)
        efficiency = math.floor((full_acc * 100) ** 3 / avg_len)

        tqdm_loop.set_postfix({'Accuracy': full_acc, 'Half Accuracy': half_acc, 'Avg retrieved length': str(avg_len),
                               'Efficiency': str(efficiency)})

    total_time = time.time() - t1
    average_precision /= counter

    print0()
    print0(f'Test Results for k = {ragConfig.top_k} :')
    print0(f'Average Precision : {average_precision}')
    print0(f'Full Accuracy (found both) : {foundBoth} out of {counter} = {full_acc} ')
    print0(f'Half Accuracy (found at least one)  : {foundOne} out of {counter} = {half_acc} ')
    print0(f'Length Accuracy Efficiency : {efficiency} ')
    print0(f'Average retrieved length : {avg_len}')

    if foundBoth > 0:
        print0(f'Average correct retrieved length : {math.floor(correct_retrieved_length / foundBoth)}')
    else:
        print0('Average correct retrieved length : N/A (no correct retrievals)')
    print0(f'Total time : {total_time}')

    print0(f'Total time : {total_time}')

    if Constants.use_wandb:
        wandb.log({
            'k': ragConfig.top_k,
            'chunk_size': rag.chunk_size,
            'Full Accuracy': full_acc,
            'Half Accuracy': half_acc,
            'Avg retrieved length': avg_len,
            'Efficiency': efficiency,
            'Average Precision': average_precision,
            'Total time': total_time,
            'corpus size': len(rag.ChunksList),
            'time to encode corpus': rag.time_to_encode_corpus,
        })

    return {
        "avg_len": avg_len,
        "efficiency": efficiency,
        "full_accuracy": full_acc,
        "half_accuracy": half_acc,
    }

def judged(rag:MetaRAG,ragConfig:MetaRagConfig,verbose=False):
    t1 = time.time()

    correct = 0
    incorrect = 0
    invalidJudgment = 0

    counter = 0

    all_llm_answer_len = 0
    # correct_retrieved_length = 0

    tqdm_loop = tqdm(ragConfig.qas.iterrows(), total=len(ragConfig.qas), desc="Evaluating",disable=not ragConfig.fine_grain_progressbar)
    if Constants.mute:tqdm_loop.disable=True

    for row in tqdm_loop:
        counter += 1
        query = row[1]['question']

        if ragConfig.multi_hop:
            answer1 = row[1]['answer1']
            answer2 = row[1]['answer2']
            ground_truth = f"{answer1}\n\n{answer2}"
        else:
            ground_truth = row[1]['answer']

        llm_answer = rag.ask(query, include_prompt=False, top_k=ragConfig.top_k, useTool=ragConfig.useTool)

        judgement = ragConfig.judge.judge(question=query, ground_truth=ground_truth, answer=llm_answer)

        llm_answer_len = token_len(llm_answer)

        all_llm_answer_len += llm_answer_len

        if "yes" in judgement.lower():
            correct += 1
        elif "no" in judgement.lower() and "Error" not in llm_answer:
            incorrect += 1
        else:
            invalidJudgment += 1

        if verbose:
            print0(f'query : {query}\n ground_truth : {ground_truth}\n answer : {llm_answer}\n judgement : {judgement}')

        correct_perc = correct / counter
        incorrect_perc = incorrect / counter
        invalidJudgment_perc = invalidJudgment / counter

        avg_len = math.floor(all_llm_answer_len / counter)
        efficiency = math.floor((correct_perc * 100) ** 3 / avg_len)

        tqdm_loop.set_postfix(
            {'Correct': correct_perc, 'Incorrect': incorrect_perc, 'Invalid Judgment': invalidJudgment_perc,
             'Avg retrieved length': str(avg_len),
             'Efficiency': str(efficiency)})

    total_time = time.time() - t1
    print0()
    print0(f'Test Results for k = {ragConfig.top_k} :')
    print0(f'Correct : {correct} out of {counter} = {correct_perc} ')
    print0(f'Incorrect : {incorrect} out of {counter} = {incorrect_perc} ')
    print0(f'Invalid Judgment : {invalidJudgment} out of {counter} = {invalidJudgment_perc} ')
    print0(f'Length Accuracy Efficiency : {efficiency} ')
    print0(f'Average retrieved length : {avg_len}')
    print0(f'Total time : {total_time}')

    if Constants.use_wandb:
        wandb.log({
            'k': ragConfig.top_k,
            'chunk_size': rag.chunk_size,
            'Correct': correct_perc,
            'Incorrect': incorrect_perc,
            'Invalid Judgment': invalidJudgment_perc,
            'Avg retrieved length': avg_len,
            'Efficiency': efficiency,
            'Total time': total_time,
            'corpus size': len(rag.ChunksList),
            'time to encode corpus': rag.time_to_encode_corpus,
            'Correct to Valid ratio': correct / (correct + incorrect),
        })

    return {
        "avg_len": avg_len,
        "efficiency": efficiency,
        "correct_perc": correct_perc,
        "incorrect_perc": incorrect_perc,
        "invalid_judgment_perc": invalidJudgment_perc,
    }

def test_retrival(ragConfig: MetaRagConfig, rag:MetaRAG=None, manage_wandb=True):
    print0("-----------------------------")
    if ragConfig.fine_grain_progressbar:print(f"\nNEW TEST multi_hop:{ragConfig.multi_hop} for encoder : {ragConfig.encoder.model_name}")
    print0("-----------------------------")

    if Constants.use_wandb and manage_wandb:
        if ragConfig.project_name is None or ragConfig.run_name is None:
            print("You need to provide wandb_project_name and wandb_config_name or set Constants.use_wandb to False")

        MRUtils.init_wandb(project_name=ragConfig.project_name, run_name=ragConfig.run_name,
                           config=ragConfig.toDict(),group=ragConfig.wandb_group)

    if rag is None:
        rag = MetaRAG(encoder_model=ragConfig.encoder, llm=ragConfig.llm)
        reset_All = True
    else:
        reset_All = False

    rag.apply_config(ragConfig)

    if reset_All:
        rag.add_corpus(ragConfig.contexts)

    if verbose_run: rag.report()

    if (ragConfig.judge is not None and ragConfig.llm is None) or (
            ragConfig.judge is None and ragConfig.llm is not None):
        print("You need to provide both judge and llm or none of them")
        return


    if ragConfig.llm is not None and ragConfig.judge is not None:
        res = judged(rag,ragConfig)
    else:
        if ragConfig.multi_hop:
            res = full_test_find_in_multiHop(rag,ragConfig)
        else:
            res = full_test_find_in(rag,ragConfig)



    gc.collect()

    return rag, res


def print0(s="\n"):
    if verbose_run: print(s)
